/*
 * newfs.c - NILFS tool, build a NILFS file system
 *
 * Copyright (C) 2005 Nippon Telegraph and Telephone Corporation.
 *
 * This file is part of NILFS.
 *
 * NILFS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * NILFS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NILFS; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * newfs.c,v 1.6 2005/09/22 02:46:30 amagai Exp
 *
 * Written by Hisashi Hifumi
 */

#define _FILE_OFFSET_BITS 64
//#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define __USE_FILE_OFFSET64

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <uuid/uuid.h>
#include <string.h>
#include <asm/page.h>
#include "newfs.h"

#define NILFS_DISKHDR_SIZE 	4096				/*HDD header(MBR+superblock)*/
#define DEFAULT_BLOCKSIZE_LOG	2				/* blocksize = 1024 * 2^2 bytes*/
#define DEFAULT_BLOCKSIZE_BYTE	(1<<(DEFAULT_BLOCKSIZE_LOG+10))
#define DEF_BLKS_PER_SEG	1024				/* default blocks per segment */
#define DEFAULT_CHKINT		15552000			/*default check interval 180days*/

#define usage_size	sizeof(struct segusage)
#define usageh_size	sizeof(struct segusage_hdr)

/*
 * linux/fs.h
 * File types
 *
 * NOTE! These match bits 12..15 of stat.st_mode
 * (ie "(i_mode >> 12) & 15").
 */
#define DT_UNKNOWN	0
#define DT_FIFO		1
#define DT_CHR		2
#define DT_DIR		4
#define DT_BLK		6
#define DT_REG		8
#define DT_LNK		10
#define DT_SOCK		12
#define DT_WHT		14

#define MOUNTS		"/etc/mtab"
#define BUFSIZE		20
struct nilfs_super_block default_sb = {
	.s_rev_level		= NILFS_CURRENT_REV,
	.s_minor_rev_level	= NILFS_MINOR_REV,
	.s_magic		= NILFS_SUPER_MAGIC,
	.s_bytes		= NILFS_SB_BYTES,
	.s_crc_seed		= 0,
	.s_sum			= 0,
	.s_nsegment		= 0,
	.s_dev_size		= 0,
	.s_first_data_block	= 0,
	.s_log_block_size	= 0,
	.s_blocks_per_segment	= 0,
	.s_r_segments_count	= 1,
	.s_last_segment		= 0,
	.s_last_seq		= 0,
	.s_free_blocks_count	= 0,
	.s_ctime		= 0,
	.s_mtime 		= 0,
	.s_wtime		= 0,
	.s_mnt_count		= 0,
	.s_max_mnt_count	= NILFS_DFL_MAX_MNT_COUNT,
	.s_state		= NILFS_VALID_FS,
	.s_errors		= 1,
	.s_lastcheck		= 0,
	.s_checkinterval	= DEFAULT_CHKINT,
	.s_creator_os		= 0,
	.s_def_resuid		= 0,
	.s_def_resgid		= 0,
	.s_first_ino		= NILFS_USER_INO,
	.s_inode_size		= sizeof(struct nilfs_inode),
	.s_uuid			= "\0",
	.s_volume_name		= "\0",
	.s_last_mounted		= "\0",
	.s_c_interval		= 0,
	.s_c_block_max		= 0,
};	/*default super block initial values */

struct nilfs_dir_entry root_dentry1 = {
	.inode			= NILFS_ROOT_INO,
	.rec_len		= 16,
	.name_len		= 1,
	.file_type		= NILFS_FT_DIR,
	.name			= ".",
};

struct nilfs_dir_entry root_dentry2 = {
	.inode			= NILFS_ROOT_INO,
	.rec_len		= 16,
	.name_len		= 2,
	.file_type		= NILFS_FT_DIR,
	.name			= "..",
};

struct nilfs_inode_hdr root_ihdr = {
	.ih_ino			= 0,
	.ih_flags		= 0,
	.ih_nfree		= 0,
	.ih_free		= NILFS_USER_INO,
};

struct nilfs_inode root_inode = {
	.i_blocks		= 1,
	.i_size			= 0,
	.i_ctime		= 0,
	.i_mtime		= 0,
	.i_dtime		= 0,
	.i_uid			= 0,
	.i_gid			= 0,
	.i_mode			= (DT_DIR << 12) | 0755,
	.i_links_count		= 2,
	.i_flags		= 0,
	.i_version		= 1,
	.i_block_root		= 0,
	.i_generation		= 0,
	.i_file_acl		= 0,
	.i_dir_acl		= 0,
};

struct nilfs_inode empty_inode = {
	.i_blocks		= 0,
	.i_size			= 0,
	.i_ctime		= 0,
	.i_mtime		= 0,
	.i_dtime		= 0,
	.i_uid			= 0,
	.i_gid			= 0,
	.i_mode			= 0,
	.i_links_count		= 0,
	.i_flags		= NILFS_INODE_NEW,
	.i_version		= 0,
	.i_block_root		= 0,
	.i_generation		= 0,
	.i_file_acl		= 0,
	.i_dir_acl		= 0,
};

struct nilfs_btree_node root_btree = {
	.bn_level		= 0,
	.bn_nkeys		= 1,
	.bn_reserved		= {0, 0, 0},
};

struct nilfs_seg_summary seg0_sum = {
	.ss_datasum		= 0,
	.ss_sumsum		= 0,
	.ss_seq			= 0,
	.ss_magic		= NILFS_SEGSUM_MAGIC,
	.ss_nblocks		= 0,
	.ss_nfinfo		= 1,
	.ss_nfbinfo		= 1,
	.ss_nfblk		= 1,
	.ss_nfbblk		= 1,
	.ss_niblk		= 1,
	.ss_nibblk		= 1,
	.ss_create		= 0,
	.ss_prev		= 0,
	.ss_flags		= SS_LOGBGN | SS_LOGEND | SS_FJCP,
};

struct nilfs_checkpoint _seg0_checkpoint = {
	.cp_sum			= 0,
	.cp_flags		= NILFS_CP_MAJOR,
	.cp_bytes		= NILFS_CP_BYTES,
	.cp_inode_root		= 0,
	.cp_nblk_lseg		= 0,
	.cp_inodes_count	= 1,
	.cp_blocks_count	= 0,
};

void sb_to_le(struct nilfs_super_block *);
void sum_to_le(struct nilfs_seg_summary *);
void dentry_to_le(struct nilfs_dir_entry *);
void ihdr_to_le(struct nilfs_inode_hdr *);
void inode_to_le(struct nilfs_inode *);
void btree_to_le(struct nilfs_btree_node *);
void checkpoint_to_le(struct nilfs_checkpoint *);
void segusage_hdr_to_le(struct segusage_hdr *);
void segusage_to_le(struct segusage *);
void usage();
void show_version();
void blksz_err();
void segsize_err();
void seek_err(__u32 blocknum, int blocksize);
void too_small_dev(__u64 dev_size, __u64 required_size);
void cannot_open_device(char *);
void cannot_get_devsize(char *);
void cannot_write_device(int);
void memalloc_err();
void disk_scan(char *, int, int, int);
void check_mount(int, char *);
char *copy_to_buffer(char *, char *, int);
void write_disk(char *, char *, char *, int, struct nilfs_super_block *, int, int, int);
void write_superblock_replica(struct nilfs_super_block *, char *, int, int);
void write_segusage_table(struct nilfs_super_block *, int, int);
char *make_segusage_table(struct nilfs_super_block *, int, __u32, unsigned long, int, struct segusage_hdr *, __u32);
unsigned long nilfs_segchunk_seg_len(long, int, unsigned long);
unsigned long nilfs_segchunk_seg_num(long, int, unsigned long);
unsigned long nilfs_susage_blk_len(long, int, unsigned long);
unsigned long long nilfs_susage_blk_num(int, long, int, unsigned long, unsigned long);
unsigned long nilfs_seg_segchunk_num(long, int, unsigned long);
unsigned int  nilfs_seg_blk_len(unsigned long, int, unsigned long, unsigned long);
void init_segusage_hdr(struct segusage_hdr *, struct nilfs_super_block *);
extern char *optarg;
extern int optind;
char *progname = "mkfs.nilfs";

int main(int argc, char *argv[])
{
	char *device, *fs_hdr, *segment0,
		*segptr, *chkpnt_ptr, volume_label[16];
	int fd, err, c, n, show_version_only = 0, quiet = 0, cflag = 0, nflag = 0;
	int blocksize, blksz_log;
	int segsize = 0, block_per_seg;
	unsigned int nblk_pseg;
	time_t t;
	__u32 crc_seed;
	__u64 dev_size = 0;
	__le64	blocks[1];
	nilfs_btree_key_t btree_key_dir, btree_key_inode;
	nilfs_btree_ptr_t btree_ptr_dir, btree_ptr_inode;
	struct nilfs_super_block *sb;
	struct nilfs_dir_entry *_root_dentry1, *_root_dentry2;
	struct nilfs_seg_summary *seg0_summary;
	struct nilfs_finfo seg0_finfo;
	struct nilfs_fbinfo seg0_fbinfo;
	struct nilfs_iinfo seg0_iinfo;
	struct nilfs_inode_hdr *root_ino_hdr;
	struct nilfs_inode *root_i, *empty_i;
	struct nilfs_btree_node *_root_btree;
	struct nilfs_checkpoint *seg0_checkpoint;

	blocksize = DEFAULT_BLOCKSIZE_BYTE;
	blksz_log = DEFAULT_BLOCKSIZE_LOG;
	block_per_seg = DEF_BLKS_PER_SEG;

	while ((c = getopt (argc, argv, "b:B:cL:nq:V")) != EOF)
		switch (c) {
			case 'b':
				blocksize = atoi(optarg);
				if (blocksize <= PAGE_SIZE && blocksize >= 1024
					&& ((blocksize - 1) & blocksize) == 0) {
					blksz_log = blocksize >> 10;
					for (n = 0; blksz_log > 1; blksz_log = blksz_log >> 1)
						n++;
					blksz_log = n;
				} else
					blksz_err();
				break;
			case 'B':
				block_per_seg = atoi(optarg);
				break;
			case 'c':
				cflag++;
				break;
			case 'L':
				strncpy(volume_label, optarg, 16);
				break;
			case 'n':
				nflag++;
				break;
			case 'q':
				quiet++;
				break;
			case 'V':
				show_version_only++;
				break;
			default:
				usage();
               }
	if ((optind == argc) && !show_version_only)
		usage();

	if (show_version_only) {
		show_version();
		exit(0);
	}

	if (block_per_seg >= NILFS_SEG_MIN_BLOCKS && ((block_per_seg-1) & block_per_seg) == 0)
		segsize = block_per_seg * blocksize;
	else
		segsize_err();

        if (argc > 0) {
                char *cp = strrchr(argv[0], '/');

                progname = (cp ? cp + 1 : argv[0]);
        }

	device = argv[optind];

	if (cflag)
		disk_scan(device, blocksize, quiet, cflag);

	fd = open(device, O_RDWR);
	if (fd < 0)
		cannot_open_device(device);

	check_mount(fd, device);

	fs_hdr = (char *) malloc((blocksize >= NILFS_DISKHDR_SIZE) ? blocksize : NILFS_DISKHDR_SIZE);
	segment0 = (char *) malloc(blocksize * 20);

	if (fs_hdr == NULL || segment0 == NULL)
		memalloc_err();

	/* start super block initialization */

	sb = &default_sb;

	/* BLKGETSIZE64 return device size in byte */
	err = ioctl(fd, BLKGETSIZE64, &dev_size);
 	if (err)
		cannot_get_devsize(device);

	sb->s_dev_size = dev_size;		/*device size in bytes*/
	sb->s_first_data_block = (blocksize >= NILFS_DISKHDR_SIZE) ? 1 : (NILFS_DISKHDR_SIZE/blocksize);
	sb->s_last_segment = sb->s_first_data_block;
	sb->s_nsegment = dev_size / segsize;
	if (sb->s_nsegment < 2)
		too_small_dev(dev_size, 2 * segsize);

	sb->s_blocks_per_segment = block_per_seg;
	sb->s_log_block_size = blksz_log;
	t=time(&t);
	srand48(t);
	crc_seed = (__u32)mrand48();

	sb->s_crc_seed = crc_seed;
	sb->s_wtime = t;                /* Write time */
	sb->s_ctime = t;
	sb->s_lastcheck = t;
	uuid_generate(sb->s_uuid);	/* set uuid using libuuid */
	memcpy(sb->s_volume_name, volume_label, 16);

	/*segment0 summary , FINFO */
	seg0_summary = &seg0_sum;
	seg0_summary->ss_create = t;
	sum_to_le(seg0_summary);

	seg0_finfo.fi_nblocks = __cpu_to_le32(1);
	seg0_finfo.fi_version = __cpu_to_le32(1);
	seg0_finfo.fi_ino = __cpu_to_le64(NILFS_ROOT_INO);
	blocks[0] = __cpu_to_le64(0);

	seg0_fbinfo.fbi_nblocks = __cpu_to_le32(1);
	seg0_fbinfo.fbi_ino = __cpu_to_le64(NILFS_ROOT_INO);

	seg0_iinfo.ii_ino = __cpu_to_le64(NILFS_ROOT_INO);

	/* data block (dentry of root directory) */
	_root_dentry1 = &root_dentry1;
	_root_dentry2 = &root_dentry2;
	_root_dentry2->rec_len = blocksize - _root_dentry2->rec_len;
	dentry_to_le(_root_dentry1);
	dentry_to_le(_root_dentry2);

	/* inode header */

	root_ino_hdr = &root_ihdr;
	root_ino_hdr->ih_nfree = (blocksize - sizeof(struct nilfs_inode_hdr)) / sizeof(struct nilfs_inode) - NILFS_USER_INO;
	ihdr_to_le(root_ino_hdr);

	/* inode block */
	root_i = &root_inode;
	root_i->i_ctime = t;
	root_i->i_mtime = t;
	root_i->i_size = blocksize;
	inode_to_le(root_i);

	empty_i = &empty_inode;
	inode_to_le(empty_i);

	/* b-tree */

	_root_btree = &root_btree;
	btree_to_le(_root_btree);
	btree_key_inode = __cpu_to_le64(NILFS_ROOT_INO /
		((blocksize - sizeof(struct nilfs_inode_hdr)) /
		sizeof(struct nilfs_inode)));

	/* checkpoint */

	seg0_checkpoint = &_seg0_checkpoint;
	seg0_checkpoint->cp_blocks_count = dev_size >> (blksz_log + 10);
	checkpoint_to_le(seg0_checkpoint);

	/*build segment0 data image written to disk */
	memset(segment0, 0, blocksize * 20);
	segptr = segment0;
	segptr = copy_to_buffer(segptr, (char *)seg0_summary, sizeof(struct nilfs_seg_summary));
	segptr = copy_to_buffer(segptr, (char *)&seg0_finfo, sizeof(struct nilfs_finfo));
	segptr = copy_to_buffer(segptr, (char *)blocks, sizeof(__le64));
	segptr = copy_to_buffer(segptr, (char *)&seg0_fbinfo, sizeof(struct nilfs_fbinfo));
	segptr = copy_to_buffer(segptr, (char *)&seg0_iinfo, sizeof(struct nilfs_iinfo));

	/* attach data blocks (root directory) */
	segptr = MOVE_TO_NEXT_BLOCK(segment0, segptr, blocksize); /* move to next block*/
	btree_key_dir = __cpu_to_le64(0);
	btree_ptr_dir = __cpu_to_le64(sb->s_first_data_block + (segptr - segment0) / blocksize);

	segptr = copy_to_buffer(segptr, (char *)_root_dentry1, __le16_to_cpu(_root_dentry1->rec_len));
	segptr = copy_to_buffer(segptr, (char *)_root_dentry2,
		min(__le16_to_cpu(_root_dentry2->rec_len), sizeof(struct nilfs_dir_entry)));

	/* attach data block B-Tree nodes */

	segptr = MOVE_TO_NEXT_BLOCK(segment0, segptr, blocksize);
	root_i->i_block_root = __cpu_to_le64(sb->s_first_data_block +
				(segptr - segment0) / blocksize);

	segptr = copy_to_buffer(segptr, (char *)_root_btree, sizeof(struct nilfs_btree_node));
	memcpy(segptr, (char *)&btree_key_dir, sizeof(nilfs_btree_key_t));  /*key*/
	segptr += (blocksize - sizeof(struct nilfs_btree_node)) / 2;

	segptr = copy_to_buffer(segptr, (char *)&btree_ptr_dir, sizeof(nilfs_btree_ptr_t));  /*ptr */

	/* attach inode blocks */
	segptr = MOVE_TO_NEXT_BLOCK(segment0, segptr, blocksize);
	btree_ptr_inode = __cpu_to_le64(sb->s_first_data_block + (segptr - segment0) / blocksize);

	segptr = copy_to_buffer(segptr, (char *)root_ino_hdr, sizeof(struct nilfs_inode_hdr));
	/* segptr += sizeof(struct nilfs_inode) * NILFS_ROOT_INO;   unused inodes space */
	segptr = copy_to_buffer(segptr, (char *)empty_i, sizeof(struct nilfs_inode));
	segptr = copy_to_buffer(segptr, (char *)empty_i, sizeof(struct nilfs_inode));

	segptr = copy_to_buffer(segptr, (char *)root_i, sizeof(struct nilfs_inode));
	{
		int nfree = (blocksize - sizeof(struct nilfs_inode_hdr)) / sizeof(struct nilfs_inode) - NILFS_USER_INO;
		for (; 0 < nfree; nfree--) /* fill inode block by empty_inode */
			segptr = copy_to_buffer(segptr, (char *)empty_i, sizeof(struct nilfs_inode));
	}
	/* attach inode B-Tree nodes */
	segptr = MOVE_TO_NEXT_BLOCK(segment0, segptr, blocksize);
	seg0_checkpoint->cp_inode_root = __cpu_to_le64(sb->s_first_data_block
					+ (segptr - segment0) / blocksize);
						/* current inode root */

	segptr = copy_to_buffer(segptr, (char *)_root_btree, sizeof(struct nilfs_btree_node));
	memcpy(segptr, (char *)&btree_key_inode, sizeof(nilfs_btree_key_t));  /*key*/
	segptr += (blocksize - sizeof(struct nilfs_btree_node)) / 2;

	segptr = copy_to_buffer(segptr, (char *)&btree_ptr_inode, sizeof(nilfs_btree_ptr_t));  /*ptr */

	/* attach checkpoint */
	segptr = MOVE_TO_NEXT_BLOCK(segment0, segptr, blocksize);
	chkpnt_ptr = segptr;
	segptr = copy_to_buffer(segptr, (char *)seg0_checkpoint, sizeof(struct nilfs_checkpoint));

	segptr = MOVE_TO_NEXT_BLOCK(segment0, segptr, blocksize);
	sb->s_free_blocks_count = (dev_size >> (blksz_log + 10))
		- sb->s_first_data_block - (segptr - segment0) / blocksize;

				/* freeblocks = total blocks - (MBR+superblock) - first written segment */

	nblk_pseg = (segptr - segment0) / blocksize;
	seg0_summary->ss_nblocks = __cpu_to_le16(nblk_pseg);
	memcpy(segment0 + sizeof(__le32) * 2 + sizeof(__le64) + sizeof(__le16),
		 (char *)&(seg0_summary->ss_nblocks), sizeof(__le16));
				/* number of blocks in segment */

	/* filling additional information in checkpoint */
	seg0_checkpoint->cp_nblk_lseg = __cpu_to_le64(nblk_pseg);
	((struct nilfs_checkpoint *)chkpnt_ptr)->cp_nblk_lseg = seg0_checkpoint->cp_nblk_lseg;

	/*segment summary checksum */
	seg0_summary->ss_sumsum
			= crc32c_le(crc_seed, segment0 + sizeof(__le32) * 2,
				sizeof(struct nilfs_seg_summary) + sizeof(struct nilfs_finfo)
				+ sizeof(__le64) + sizeof(struct nilfs_fbinfo)
				+ sizeof(struct nilfs_iinfo) - sizeof(__le32) * 2);
	seg0_summary->ss_sumsum = __cpu_to_le32(seg0_summary->ss_sumsum);
	memcpy(segment0 + sizeof(__le32), (char *)&(seg0_summary->ss_sumsum), sizeof(__le32));

	/* checkpoint checksum */
	seg0_checkpoint->cp_sum
		= crc32c_le(crc_seed, chkpnt_ptr + sizeof(__le32), NILFS_CP_BYTES - sizeof(__le32));
	seg0_checkpoint->cp_sum = __cpu_to_le32(seg0_checkpoint->cp_sum);
	memcpy(chkpnt_ptr, (char *)&(seg0_checkpoint->cp_sum), sizeof(__le32));

	/* segment0 checksum */
	seg0_summary->ss_datasum
			= crc32c_le(crc_seed, segment0 + sizeof(__le32),
				segptr - segment0 - sizeof(__le32));

	seg0_summary->ss_datasum = __cpu_to_le32(seg0_summary->ss_datasum);
	memcpy(segment0, (char *)&(seg0_summary->ss_datasum), sizeof(__le32));

	/*start writing disk */
	write_disk(fs_hdr, segment0, device, blocksize, sb, fd, nflag, quiet);

	exit(0);
}

void sb_to_le (struct nilfs_super_block *sb)
{
	sb->s_rev_level = __cpu_to_le32(sb->s_rev_level);
	sb->s_minor_rev_level = __cpu_to_le16(sb->s_minor_rev_level);
	sb->s_magic = __cpu_to_le16(sb->s_magic);
	sb->s_bytes = __cpu_to_le16(sb->s_bytes);
	sb->s_crc_seed = __cpu_to_le32(sb->s_crc_seed);
	sb->s_nsegment = __cpu_to_le32(sb->s_nsegment);
	sb->s_dev_size = __cpu_to_le64(sb->s_dev_size);
	sb->s_first_data_block = __cpu_to_le64(sb->s_first_data_block);
	sb->s_log_block_size = __cpu_to_le32(sb->s_log_block_size);
	sb->s_blocks_per_segment = __cpu_to_le32(sb->s_blocks_per_segment);
	sb->s_r_segments_count = __cpu_to_le32(sb->s_r_segments_count);
	sb->s_last_segment = __cpu_to_le64(sb->s_last_segment);
	sb->s_last_seq = __cpu_to_le64(sb->s_last_seq);
	sb->s_free_blocks_count = __cpu_to_le64(sb->s_free_blocks_count);
	sb->s_ctime = __cpu_to_le64(sb->s_ctime);
	sb->s_mtime = __cpu_to_le64(sb->s_mtime);
	sb->s_wtime = __cpu_to_le64(sb->s_wtime);
	sb->s_mnt_count = __cpu_to_le16(sb->s_mnt_count);
	sb->s_max_mnt_count = __cpu_to_le16(sb->s_max_mnt_count);
	sb->s_state = __cpu_to_le16(sb->s_state);
	sb->s_errors = __cpu_to_le16(sb->s_errors);
	sb->s_lastcheck = __cpu_to_le64(sb->s_lastcheck);
	sb->s_checkinterval = __cpu_to_le32(sb->s_checkinterval);
	sb->s_creator_os = __cpu_to_le32(sb->s_creator_os);
	sb->s_def_resuid = __cpu_to_le16(sb->s_def_resuid);
 	sb->s_def_resgid = __cpu_to_le16(sb->s_def_resgid);
	sb->s_first_ino = __cpu_to_le32(sb->s_first_ino);
  	sb->s_inode_size = __cpu_to_le16(sb->s_inode_size);
	sb->s_c_interval = __cpu_to_le32(sb->s_c_interval);
	sb->s_c_block_max = __cpu_to_le32(sb->s_c_block_max);
}

void sum_to_le(struct nilfs_seg_summary *seg0_summary)
{
	seg0_summary->ss_seq = __cpu_to_le64(seg0_summary->ss_seq);
	seg0_summary->ss_magic = __cpu_to_le16(seg0_summary->ss_magic);
	seg0_summary->ss_nblocks = __cpu_to_le16(seg0_summary->ss_nblocks);
	seg0_summary->ss_nfinfo = __cpu_to_le16(seg0_summary->ss_nfinfo);
	seg0_summary->ss_nfbinfo = __cpu_to_le16(seg0_summary->ss_nfbinfo);
	seg0_summary->ss_nfblk = __cpu_to_le16(seg0_summary->ss_nfblk);
	seg0_summary->ss_nfbblk = __cpu_to_le16(seg0_summary->ss_nfbblk);
	seg0_summary->ss_niblk = __cpu_to_le16(seg0_summary->ss_niblk);
	seg0_summary->ss_nibblk = __cpu_to_le16(seg0_summary->ss_nibblk);
	seg0_summary->ss_create = __cpu_to_le64(seg0_summary->ss_create);
	seg0_summary->ss_prev = __cpu_to_le16(seg0_summary->ss_prev);
	seg0_summary->ss_flags = __cpu_to_le16(seg0_summary->ss_flags);
}

void dentry_to_le(struct nilfs_dir_entry *_root_dentry)
{
	_root_dentry->inode = __cpu_to_le64(_root_dentry->inode);
	_root_dentry->rec_len = __cpu_to_le16(_root_dentry->rec_len);
}

void ihdr_to_le(struct nilfs_inode_hdr *root_ino_hdr)
{
	root_ino_hdr->ih_ino = __cpu_to_le64(root_ino_hdr->ih_ino);
	root_ino_hdr->ih_flags = __cpu_to_le16(root_ino_hdr->ih_flags);
	root_ino_hdr->ih_nfree = __cpu_to_le16(root_ino_hdr->ih_nfree);
	root_ino_hdr->ih_free = __cpu_to_le32(root_ino_hdr->ih_free);
}

void inode_to_le(struct nilfs_inode *root_i)
{
	root_i->i_blocks = __cpu_to_le64(root_i->i_blocks);
	root_i->i_size = __cpu_to_le64(root_i->i_size);
	root_i->i_ctime = __cpu_to_le64(root_i->i_ctime);
	root_i->i_mtime = __cpu_to_le64(root_i->i_mtime);
	root_i->i_dtime = __cpu_to_le64(root_i->i_dtime);
	root_i->i_uid = __cpu_to_le32(root_i->i_uid);
	root_i->i_gid = __cpu_to_le32(root_i->i_gid);
	root_i->i_mode = __cpu_to_le16(root_i->i_mode);
	root_i->i_links_count = __cpu_to_le16(root_i->i_links_count);
	root_i->i_flags = __cpu_to_le32(root_i->i_flags);
	root_i->i_version = __cpu_to_le32(root_i->i_version);
	root_i->i_block_root = __cpu_to_le64(root_i->i_block_root);
	root_i->i_generation = __cpu_to_le32(root_i->i_generation);
	root_i->i_file_acl = __cpu_to_le32(root_i->i_file_acl);
	root_i->i_dir_acl = __cpu_to_le32(root_i->i_dir_acl);
}

void btree_to_le(struct nilfs_btree_node *_root_btree)
{
	_root_btree->bn_level = __cpu_to_le16(_root_btree->bn_level);
	_root_btree->bn_nkeys = __cpu_to_le16(_root_btree->bn_nkeys);
}

void checkpoint_to_le(struct nilfs_checkpoint *seg0_checkpoint)
{
	seg0_checkpoint->cp_flags = __cpu_to_le16(seg0_checkpoint->cp_flags);
	seg0_checkpoint->cp_bytes = __cpu_to_le16(seg0_checkpoint->cp_bytes);
	seg0_checkpoint->cp_inode_root = __cpu_to_le64(seg0_checkpoint->cp_inode_root);
	seg0_checkpoint->cp_nblk_lseg = __cpu_to_le64(seg0_checkpoint->cp_nblk_lseg);
	seg0_checkpoint->cp_inodes_count = __cpu_to_le64(seg0_checkpoint->cp_inodes_count);
	seg0_checkpoint->cp_blocks_count = __cpu_to_le64(seg0_checkpoint->cp_blocks_count);
}

void segusage_hdr_to_le(struct segusage_hdr *segusage_hdr_tmp)
{
	segusage_hdr_tmp->suh_free_blocks_count
			= __cpu_to_le64(segusage_hdr_tmp->suh_free_blocks_count);
	segusage_hdr_tmp->suh_last_segment
			= __cpu_to_le64(segusage_hdr_tmp->suh_last_segment);
	segusage_hdr_tmp->suh_wtime = __cpu_to_le64(segusage_hdr_tmp->suh_wtime);
	segusage_hdr_tmp->suh_tm = __cpu_to_le64(segusage_hdr_tmp->suh_tm);
	segusage_hdr_tmp->suh_aux = __cpu_to_le32(segusage_hdr_tmp->suh_aux);
	segusage_hdr_tmp->suh_sum = __cpu_to_le32(segusage_hdr_tmp->suh_sum);
}

void segusage_to_le(struct segusage *segusage_tmp)
{
	segusage_tmp->su_prev_full_seg = __cpu_to_le32(segusage_tmp->su_prev_full_seg);
	segusage_tmp->su_next_full_seg = __cpu_to_le32(segusage_tmp->su_next_full_seg);
	segusage_tmp->su_lastmod = __cpu_to_le64(segusage_tmp->su_lastmod);
	segusage_tmp->su_aux = __cpu_to_le16(segusage_tmp->su_aux);
	segusage_tmp->su_flags = __cpu_to_le16(segusage_tmp->su_flags);
}

void usage()
{
	fprintf(stderr, "usage:%s [-b block-size] [-B blocks per segment] [-c] \n[-L volume-label] [-q] [-r revision level] [-V] device \n", progname);
	exit(1);
}

void show_version()
{
	int current = NILFS_CURRENT_REV, minor = NILFS_MINOR_REV;
	printf("%s ver %d.%d\n", progname, current, minor);
}

void blksz_err()
{
	show_version();
	fprintf(stderr, "Error: Bad blocksize.\n");
	exit(1);
}

void segsize_err()
{
	show_version();
	fprintf(stderr, "Error: Bad number of blocks per segment.\n");
	exit(1);
}

void too_small_dev(__u64 dev_size, __u64 required_size)
{
	show_version();
	fprintf(stderr,	"Error: Too small device.\n       "
		"device size=%llu bytes, "
		"required size=%llu bytes.\n       ",
		dev_size, required_size);
	fprintf(stderr, "Please enlarge the device, "
		"or shorten segments with -B option.\n");
	exit(1);
}

void seek_err(__u32 blocknum, int blocksize)
{
	show_version();
	fprintf(stderr, "Error: can not seek to %u blknum %u!\n",
		blocknum * blocksize,
		blocknum);
	perror("");
	exit(1);
}

void cannot_open_device(char *device)
{
	show_version();
	fprintf(stderr, "can not open device! (%s)\n", device);
	exit(1);
}

void cannot_get_devsize(char *device)
{
	show_version();
	fprintf(stderr, "can not get device size! (%s)\n", device);
	exit(1);
}

void cannot_write_device(int fd)
{
	show_version();
	fprintf(stderr, "can not write device!\n");
	close(fd);
	exit(1);
}

void memalloc_err()
{
	show_version();
	fprintf(stderr, "Error: Memory Allocation Error.\n");
	exit(1);
}

void disk_scan(char *device, int blocksize, int quiet, int cflag)
{
	char buf[1024];
	sprintf(buf, "badblocks -b %d %s %s %s\n", blocksize,
		quiet ? "" : "-s", (cflag > 1) ? "-w" : "", device);

	if (!quiet)
		printf("check blocks\n");

	system(buf);
}

void check_mount(int fd, char *device)
{
	FILE *fp;
	char line[BUFSIZE];

	fp = fopen(MOUNTS, "r");
	if (fp == NULL){
		show_version();
		fprintf(stderr, "Cannot open /etc/mtab!!\n");
		close(fd);
		exit(1);
	}

	while (fgets(line, BUFSIZE, fp)!=NULL){
		if (strncmp(strtok(line, " "), device, strlen(device)) == 0) {
			show_version();
			fprintf(stderr, "%s is currently mounted. You cannot make a filesystem on this device.\n", device);
			fclose(fp);
			close(fd);
			exit(1);
		}
	}
	fclose(fp);
}

char *copy_to_buffer(char *ptr1, char *ptr2, int bytes)
{
	memcpy(ptr1, ptr2, bytes);
	ptr1 += bytes;

	return ptr1;
}

void write_disk(char *fs_hdr, char *segment0, char *device, int blocksize,
		struct nilfs_super_block *sb, int fd, int nflag, int quiet)
{
	int err;
	__u32 crc_seed = sb->s_crc_seed;

	/* build fs header image including superblock */
	memset(fs_hdr, 0, (blocksize >= NILFS_DISKHDR_SIZE) ? blocksize : NILFS_DISKHDR_SIZE);
	err = read(fd, fs_hdr, NILFS_SB_OFFSET_BYTES);
	if (err < 0) {
		show_version();
		fprintf(stderr, "can not read device!\n");
		close(fd);
		exit(1);
	}

	sb_to_le(sb);
	/* superblock checksum */
	sb->s_sum = __cpu_to_le32(crc32c_le(crc_seed, (char *)sb, NILFS_SB_BYTES));

	memcpy(fs_hdr + NILFS_SB_OFFSET_BYTES, (char *)sb, sizeof(struct nilfs_super_block));

	/* start writing to HDD*/

	/* write superblock */
	lseek(fd, 0, SEEK_SET);

	if (!nflag) {
		err = write(fd, fs_hdr, (blocksize >= NILFS_DISKHDR_SIZE) ? blocksize : NILFS_DISKHDR_SIZE);
		if (err < 0)
			cannot_write_device(fd);
	}

		/* write segment0 */
	if (!quiet) {
		show_version();
		printf("Start writing file system initial data to the device \n");
		printf("Blocksize:%d  Device:%s \n", blocksize, device);
	}

	if (!nflag) {
		err = write(fd, segment0, blocksize * 20);
		if (err < 0)
			cannot_write_device(fd);
	}

	if (!nflag) {
		write_segusage_table(sb, blocksize, fd);
		write_superblock_replica(sb, fs_hdr, blocksize, fd);
	}

	if (!nflag) {
		err = fsync(fd);
		if (err < 0)
			cannot_write_device(fd);
	}
	if (!quiet)
		fprintf(stderr, "File system initialization succeeded !! \n");

	close(fd);
}

void
write_superblock_replica(struct nilfs_super_block *sb, char *fs_hdr, int blocksize, int fd)
{
	__u32 blocknum = sb->s_blocks_per_segment;

	if (lseek64(fd, (__u64)blocknum * blocksize, SEEK_SET) < 0)
		seek_err(blocknum, blocksize);

	if (write(fd, fs_hdr, blocksize) < 0)
		cannot_write_device(fd);
}

void
write_segusage_table(struct nilfs_super_block *sb, int blocksize, int fd)
{
	__u32 usage_blocks, blocknum;
	__u32 nsegment = sb->s_nsegment;
	__u32 blocks_per_segment = sb->s_blocks_per_segment;
	struct segusage_hdr segusage_hdr_skeleton;
	char *segusage_table;
	unsigned long n_usage_b = (blocksize - usageh_size) / usage_size;
	long onchunk = (nsegment + n_usage_b - 1) / n_usage_b;
	unsigned long segnum = 0;
	int secondary, chunk, seglen;
#if 0
	{ /* for debug */
		printf("su hdr size %d su size %d\n",
		       usageh_size, usage_size);
		printf("dev_size %lld nsegment %d blocks_per_segment %d n_usage_b %lu\n",
		       sb->s_dev_size, nsegment, blocks_per_segment, n_usage_b);
	}
#endif
	onchunk = onchunk < NILFS_MAX_SU_NCHUNK ? onchunk : NILFS_MAX_SU_NCHUNK;
	init_segusage_hdr(&segusage_hdr_skeleton, sb);
	for (chunk = 0; chunk < onchunk; chunk++) {
		usage_blocks = nilfs_susage_blk_len(chunk, blocksize, nsegment);
		seglen = nilfs_segchunk_seg_len(chunk, blocksize, nsegment);
		segusage_table = make_segusage_table(sb, blocksize, chunk, segnum, seglen,
						     &segusage_hdr_skeleton, usage_blocks);
		for (secondary = 0; secondary < 2; secondary++) {
			blocknum = nilfs_susage_blk_num(secondary, chunk, blocksize, nsegment, blocks_per_segment);
#if 0
			printf("chunk %2d-%d seg %lu blk %d seg %d/%d len %d\n",
			       chunk, secondary, segnum-1, blocknum,
			       blocknum/blocks_per_segment, blocknum%blocks_per_segment,
			       usage_blocks);
#endif
			if (lseek64(fd, (__u64)blocknum * blocksize, SEEK_SET) < 0)
				seek_err(blocknum, blocksize);
			if (write(fd, segusage_table, blocksize * usage_blocks) < 0)
				cannot_write_device(fd);
		}
		segnum += seglen;
		free(segusage_table);
	}
}

char *
make_segusage_table(struct nilfs_super_block *sb, int blocksize, __u32 segchunknum,
		    unsigned long segnum, int seglen,
		    struct segusage_hdr *segusage_hdr_skeleton,
		    __u32 usage_blocks)
{
	__u32 nsegment = sb->s_nsegment;
	__u32 blocks_per_segment = sb->s_blocks_per_segment;
	unsigned long n_usage_b = (blocksize - usageh_size) / usage_size;
	__u32 checksum;
	char *ptr, *segusage_table;
	struct segusage_hdr segusage_hdr_tmp;
	struct segusage segusage_tmp;
	unsigned long sseg;
	int ns, nseg, block;

	memcpy(&segusage_hdr_tmp, segusage_hdr_skeleton, usageh_size);

	segusage_table = (char *)malloc(blocksize * usage_blocks);
	if (segusage_table == NULL)
		memalloc_err();
	memset(segusage_table, 0, blocksize * usage_blocks);

	ptr = segusage_table;
	/* build structure of segusage */
	segusage_hdr_to_le(&segusage_hdr_tmp);

	sseg = segnum;
	nseg = nilfs_segchunk_seg_len(segchunknum, blocksize, nsegment);

	block = ns = 0;
	for (; segnum < sseg + seglen; segnum++) {
		if (ns == 0)
			ptr = copy_to_buffer(ptr, (char *)&segusage_hdr_tmp, usageh_size);

		memset(&segusage_tmp, 0, usage_size);
		if (segchunknum == 0 && segnum == 0) {
			segusage_tmp.su_prev_full_seg = SU_NOSEG;
			segusage_tmp.su_flags = SEGUSE_USED|SEGUSE_ACTIVE;
		} else
			segusage_tmp.su_prev_full_seg = segnum - 1;
		if (segnum == nsegment - 1)
			segusage_tmp.su_next_full_seg = SU_NOSEG;
		else
			segusage_tmp.su_next_full_seg = segnum + 1;
		if (nilfs_seg_blk_len(segnum, blocksize, nsegment, blocks_per_segment) < blocks_per_segment)
			segusage_tmp.su_flags |= SEGUSE_SHORTLEN;
		segusage_to_le(&segusage_tmp);
		ptr = copy_to_buffer(ptr, (char *)&segusage_tmp, usage_size);

		if (++ns == n_usage_b) {
			ptr = MOVE_TO_NEXT_BLOCK(segusage_table, ptr, blocksize);
			block++;
			ns = 0;
		}
	}
	for (block = 0; block < usage_blocks; block++) {
		checksum = crc32c_le(sb->s_crc_seed, segusage_table + blocksize * block + sizeof(__le32),
				     blocksize - sizeof(__le32));
		checksum = __cpu_to_le32(checksum);
		memcpy(segusage_table + blocksize * block, (char *)&checksum, sizeof(__le32));
	}
	return segusage_table;
}

/* segment chunk number -> number of segments of the segment chunk */
unsigned long
nilfs_segchunk_seg_len(long segchunknum, int blocksize, unsigned long nsegment)
{
	unsigned long n_usage_b = (blocksize - usageh_size) / usage_size;
	long onchunk = (nsegment + n_usage_b - 1) / n_usage_b;
	long nb, nbm, s;

	if (segchunknum < 0 || NILFS_MAX_SU_NCHUNK <= segchunknum) {
		show_version();
		fprintf(stderr, "segchunknum range error\n");
	}
	if (onchunk <= NILFS_MAX_SU_NCHUNK) {
		if (onchunk <= segchunknum) {
			show_version();
			fprintf(stderr, "segchunknum range error\n");
		}
		if (segchunknum == onchunk - 1) { /* last segment chunk */
			if (nsegment % n_usage_b == 1) /* at least 2 segments */ /* type I */
				return 2;
			else
				return nsegment % n_usage_b;	/* type G */
		} else if (segchunknum == onchunk - 2 && nsegment % n_usage_b == 1)
			return n_usage_b - 1;	/* type H */
		else	/* type F */
			return n_usage_b;
	}

	nb = onchunk / NILFS_MAX_SU_NCHUNK;
	nbm = onchunk % NILFS_MAX_SU_NCHUNK;
	if (segchunknum < nbm - 1) { 	/* type A */
		return (nb + 1) * n_usage_b;
	} else if (segchunknum == nbm - 1) {	/* type B */
		if ((s = nsegment % n_usage_b))
			return nb * n_usage_b + s;
		else
			return (nb + 1) * n_usage_b;
	} else if (segchunknum == NILFS_MAX_SU_NCHUNK - 1 && nbm == 0 && (s = nsegment % n_usage_b)) { /* type E */
		return (nb - 1) * n_usage_b + s;
	} else {	/* type C, D */
		return nb * n_usage_b;
	}
}

/* segment chunk number -> segment number of beginning of the segment chunk */
unsigned long
nilfs_segchunk_seg_num(long segchunknum, int blocksize, unsigned long nsegment)
{
	unsigned long n_usage_b = (blocksize - usageh_size) / usage_size;
	long onchunk = (nsegment + n_usage_b - 1) / n_usage_b;
	long nb, nbm;

	if (segchunknum < 0 || NILFS_MAX_SU_NCHUNK <= segchunknum) {
		show_version();
		fprintf(stderr, "segchunknum range error\n");
	}
	if (onchunk <= NILFS_MAX_SU_NCHUNK) {
		if (onchunk <= segchunknum) {
			show_version();
			fprintf(stderr, "segchunknum range error\n");
		}
		if (segchunknum == onchunk - 1 && nsegment % n_usage_b == 1)
			return segchunknum * n_usage_b - 1;	/* type I */
		else	/* type F, G, H */
			return segchunknum * n_usage_b;
	}

	nb = onchunk / NILFS_MAX_SU_NCHUNK;
	nbm = onchunk % NILFS_MAX_SU_NCHUNK;
	if (segchunknum < nbm) {	/* type A, B */
		return segchunknum * (nb + 1) * n_usage_b;
	} else if (nbm == 0) {	/* type D, E */
		return segchunknum * nb * n_usage_b;
	} else {		/* type C */
		return nsegment - nb * n_usage_b * (NILFS_MAX_SU_NCHUNK - segchunknum);
	}
}

/* segment chunk number -> number of blocks for segment usage table */
unsigned long
nilfs_susage_blk_len(long segchunknum, int blocksize, unsigned long nsegment)
{
	unsigned long n_usage_b = (blocksize - usageh_size) / usage_size;
	long onchunk = (nsegment + n_usage_b - 1) / n_usage_b;

	if (segchunknum < 0 || NILFS_MAX_SU_NCHUNK <= segchunknum) {
		show_version();
		fprintf(stderr, "segchunknum range error\n");
	}
	if (onchunk <= NILFS_MAX_SU_NCHUNK) /* type F, G, H, I */
		return 1;
	else if (segchunknum < onchunk % NILFS_MAX_SU_NCHUNK) /* type A, B */
		return onchunk / NILFS_MAX_SU_NCHUNK + 1;
	else			/* type C, D, E */
		return onchunk / NILFS_MAX_SU_NCHUNK;
}

/* segment number -> segment chunk number */
unsigned long
nilfs_seg_segchunk_num(long segnum, int blocksize, unsigned long nsegment)
{
	unsigned long n_usage_b = (blocksize - usageh_size) / usage_size;
	long onchunk = (nsegment + n_usage_b - 1) / n_usage_b;
	long nb, nbm, segchunknum;

	if (segnum < 0 || nsegment <= segnum) {
		show_version();
		fprintf(stderr, "segnum range error\n");
	}
	if (onchunk <= NILFS_MAX_SU_NCHUNK) {
		if (nsegment % n_usage_b == 1 && segnum == nsegment - 2)
			segchunknum = segnum / n_usage_b + 1;	/* type I */
		else
			segchunknum = segnum / n_usage_b; 	/* type F, G, H */
	} else {
		nb = onchunk / NILFS_MAX_SU_NCHUNK;
		nbm = onchunk % NILFS_MAX_SU_NCHUNK;

		if (0 < nbm) { /* type A, B, C */
			if (nsegment - (NILFS_MAX_SU_NCHUNK - nbm) * nb * n_usage_b <= segnum) { /* type C */
				segchunknum = NILFS_MAX_SU_NCHUNK - (nsegment - segnum - 1) / (nb * n_usage_b) - 1;
			} else if (segnum < (nb + 1) * n_usage_b * (nbm - 1)) { /* type A */
				segchunknum = segnum / ((nb + 1) * n_usage_b);
			} else	{ /* type B */
				segchunknum = nbm - 1;
			}
		} else {	/* type D, E */
			segchunknum = segnum / (nb * n_usage_b);
		}
	}
	return segchunknum;
}

/* segment chunk number -> block number of segment usage table header */
unsigned long long
nilfs_susage_blk_num(int secondary, long segchunknum, int blocksize,
		     unsigned long nsegment, unsigned long blocks_per_segment)
{
	unsigned long n_usage_b = (blocksize - usageh_size) / usage_size;
	long onchunk = (nsegment + n_usage_b - 1) / n_usage_b;
	long nb, nbm;

	if (segchunknum < 0 || NILFS_MAX_SU_NCHUNK <= segchunknum) {
		show_version();
		fprintf(stderr, "segchunknum range error\n");
	}
	if (secondary) {
		long sseg = segchunknum + 1;
		long o = onchunk < NILFS_MAX_SU_NCHUNK ? onchunk : NILFS_MAX_SU_NCHUNK;

		if (sseg == o)
			sseg = 0;
		return (nilfs_segchunk_seg_num(sseg, blocksize, nsegment)
			+ nilfs_segchunk_seg_len(sseg, blocksize, nsegment) / 2)
			* blocks_per_segment - nilfs_susage_blk_len(segchunknum, blocksize, nsegment);
	}
	if (onchunk <= NILFS_MAX_SU_NCHUNK) {
		if (onchunk <= segchunknum) {
			show_version();
			fprintf(stderr, "segchunknum range error\n");
		}
		if (segchunknum == onchunk - 1)	/* type G, I */
			return blocks_per_segment * nsegment - 1;
		else if (segchunknum == onchunk - 2 && nsegment % n_usage_b == 1) /* last segment chunk has 2 segments */ /* type H */
			return (segchunknum + 1 ) * n_usage_b * blocks_per_segment - blocks_per_segment - 1;
		else	/* type F */
			return (segchunknum + 1) * n_usage_b * blocks_per_segment - 1;
	}
	nb = onchunk / NILFS_MAX_SU_NCHUNK;
	nbm = onchunk % NILFS_MAX_SU_NCHUNK;
	if (segchunknum < nbm - 1) {	/* type A */
		return n_usage_b * (nb + 1) * blocks_per_segment * (segchunknum + 1) - (nb + 1);
	} else if (segchunknum == nbm - 1) {	/* type B */
		return blocks_per_segment * nsegment
			- (n_usage_b * nb * blocks_per_segment * (NILFS_MAX_SU_NCHUNK - segchunknum - 1))
			- (nb + 1);
	} else if (nbm == 0) {
		if (segchunknum == NILFS_MAX_SU_NCHUNK - 1)	/* type E */
			return blocks_per_segment * nsegment - nb;
		else	/* type D */
			return (segchunknum + 1) * n_usage_b * nb * blocks_per_segment - nb;
	} else {	/* type C */
		return blocks_per_segment * nsegment
			- (n_usage_b * nb * blocks_per_segment * (NILFS_MAX_SU_NCHUNK - segchunknum - 1))
			- nb;
	}
}

/* segnum segment length in block */
unsigned int
nilfs_seg_blk_len(unsigned long segnum, int blocksize, unsigned long nsegment, unsigned long blocks_per_segment)
{
	unsigned long segchunknum = nilfs_seg_segchunk_num(segnum, blocksize, nsegment);
	unsigned long chunk_start = nilfs_segchunk_seg_num(segchunknum, blocksize, nsegment);
	unsigned long chunk_len = nilfs_segchunk_seg_len(segchunknum, blocksize, nsegment);
	unsigned long blklen;

	if (segnum == chunk_start + chunk_len - 1) {	/* last segment in the chunk */
		blklen = blocks_per_segment - nilfs_susage_blk_len(segchunknum, blocksize, nsegment);
	} else {
		unsigned long n_usage_b = (blocksize - usageh_size) / usage_size;
		long onchunk = (nsegment + n_usage_b - 1) / n_usage_b;
		long o = onchunk < NILFS_MAX_SU_NCHUNK ? onchunk : NILFS_MAX_SU_NCHUNK;
		long sseg = segchunknum - 1;

		if (sseg < 0)
			sseg = o - 1;
		if (segnum == chunk_start + chunk_len / 2 - 1) {
			/* secondary usage table */
			blklen = blocks_per_segment - nilfs_susage_blk_len(sseg, blocksize, nsegment);
		} else {
			blklen = blocks_per_segment;
		}
	}
	if (segnum == 0)	/* super block */
		blklen--;
	else if (segnum == 1)
		blklen--;	/* super block replica */
	return blklen;
}

void init_segusage_hdr(struct segusage_hdr *segusage_hdr_skeleton, struct nilfs_super_block *sb)
{
	segusage_hdr_skeleton->suh_sum = 0;
	segusage_hdr_skeleton->suh_aux = 0;
	segusage_hdr_skeleton->suh_free_blocks_count = sb->s_free_blocks_count;
	segusage_hdr_skeleton->suh_last_segment = sb->s_last_segment;
	segusage_hdr_skeleton->suh_wtime = sb->s_wtime;
	segusage_hdr_skeleton->suh_tm = sb->s_wtime;

}

/* Local Variables:	*/
/* eval: (c-set-style "linux")	*/
/* End:			*/
